@@ -3,15 +3,26 @@
 require_relative 'helper'
 require 'tempfile'
 
+separate_testing do
+  require_relative '../lib/rack/lint'
+  require_relative '../lib/rack/mock_request'
+end
+
 describe Rack::Lint do
-  def env(*args)
-    Rack::MockRequest.env_for("/", *args)
+  valid_app = lambda do |env|
+    [200, { "content-type" => "test/plain", "content-length" => "3" }, ["foo"]]
+  end
+
+  def env(options = {})
+    unless options.key?(:input)
+      options[:input] = String.new
+    end
+
+    Rack::MockRequest.env_for("/", options)
   end
 
   it "pass valid request" do
-    Rack::Lint.new(lambda { |env|
-                     [200, { "Content-type" => "test/plain", "Content-length" => "3" }, ["foo"]]
-                   }).call(env({})).first.must_equal 200
+    Rack::Lint.new(valid_app).call(env({})).first.must_equal 200
   end
 
   it "notice fatal errors" do
@@ -26,7 +37,6 @@ describe Rack::Lint do
     lambda { Rack::Lint.new(nil).call({}.freeze) }.must_raise(Rack::Lint::LintError).
       message.must_match(/env should not be frozen, but is/)
 
-
     lambda {
       e = env
       e.delete("REQUEST_METHOD")
@@ -41,6 +51,19 @@ describe Rack::Lint do
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/missing required key SERVER_NAME/)
 
+    lambda {
+      e = env
+      e.delete("SERVER_PROTOCOL")
+      Rack::Lint.new(nil).call(e)
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/missing required key SERVER_PROTOCOL/)
+
+    lambda {
+      e = env
+      e["SERVER_PROTOCOL"] = 'Foo'
+      Rack::Lint.new(nil).call(e)
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/env\[SERVER_PROTOCOL\] does not match HTTP/)
 
     lambda {
       Rack::Lint.new(nil).call(env("HTTP_CONTENT_TYPE" => "text/plain"))
@@ -57,11 +80,6 @@ describe Rack::Lint do
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/non-string value/)
 
-    lambda {
-      Rack::Lint.new(nil).call(env("rack.version" => "0.2"))
-    }.must_raise(Rack::Lint::LintError).
-      message.must_match(/must be an Array/)
-
     lambda {
       Rack::Lint.new(nil).call(env("rack.url_scheme" => "gopher"))
     }.must_raise(Rack::Lint::LintError).
@@ -72,6 +90,8 @@ describe Rack::Lint do
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "session [] must respond to store and []="
 
+    Rack::Lint.new(valid_app).call(env("rack.session" => {}))[0].must_equal 200
+
     lambda {
       Rack::Lint.new(nil).call(env("rack.session" => {}.freeze))
     }.must_raise(Rack::Lint::LintError).
@@ -133,11 +153,16 @@ describe Rack::Lint do
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "logger [] must respond to fatal"
 
+    def obj.fatal(*) end
+    Rack::Lint.new(valid_app).call(env("rack.logger" => obj))[0].must_equal 200
+
     lambda {
       Rack::Lint.new(nil).call(env("rack.multipart.buffer_size" => 0))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "rack.multipart.buffer_size must be an Integer > 0 if specified"
 
+    Rack::Lint.new(valid_app).call(env("rack.multipart.buffer_size" => 1))[0].must_equal 200
+
     lambda {
       Rack::Lint.new(nil).call(env("rack.multipart.tempfile_factory" => Tempfile))
     }.must_raise(Rack::Lint::LintError).
@@ -158,6 +183,21 @@ describe Rack::Lint do
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "response array has 0 elements instead of 3"
 
+    lambda {
+      Rack::Lint.new(nil).call(env("SERVER_PORT" => "howdy"))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'env[SERVER_PORT] is not an Integer'
+
+    lambda {
+      Rack::Lint.new(nil).call(env("SERVER_NAME" => "\u1234"))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "\u1234 must be a valid authority"
+
+    lambda {
+      Rack::Lint.new(nil).call(env("HTTP_HOST" => "\u1234"))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "\u1234 must be a valid authority"
+
     lambda {
       Rack::Lint.new(nil).call(env("REQUEST_METHOD" => "FUCKUP?"))
     }.must_raise(Rack::Lint::LintError).
@@ -173,11 +213,6 @@ describe Rack::Lint do
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/must start with/)
 
-    lambda {
-      Rack::Lint.new(nil).call(env("PATH_INFO" => "../foo"))
-    }.must_raise(Rack::Lint::LintError).
-      message.must_match(/must start with/)
-
     lambda {
       Rack::Lint.new(nil).call(env("CONTENT_LENGTH" => "xcii"))
     }.must_raise(Rack::Lint::LintError).
@@ -209,6 +244,16 @@ describe Rack::Lint do
       Rack::Lint.new(nil).call(env("SCRIPT_NAME" => "/"))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/cannot be .* make it ''/)
+
+    lambda {
+      Rack::Lint.new(nil).call(env("rack.response_finished" => "not a callable"))
+    }.must_raise(Rack::Lint::LintError).
+    message.must_match(/rack.response_finished must be an array of callable objects/)
+
+    lambda {
+      Rack::Lint.new(nil).call(env("rack.response_finished" => [-> (env) {}, "not a callable"]))
+    }.must_raise(Rack::Lint::LintError).
+    message.must_match(/rack.response_finished values must respond to call/)
   end
 
   it "notice input errors" do
@@ -265,32 +310,115 @@ describe Rack::Lint do
       message.must_include('response array has 4 elements instead of 3')
   end
 
+  it "accepts empty PATH_INFO" do
+    Rack::Lint.new(valid_app).call(env("PATH_INFO" => "")).first.must_equal 200
+  end
+
+  it "notices request-target asterisk form errors" do
+    # A non-empty PATH_INFO starting with something other than / has
+    # implications for Rack::Request#path and methods downstream from
+    # it. Note that RFC3875 does not actually anticipate dealing with
+    # `OPTIONS *`; that should be considered a bug in the spec.
+    Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "OPTIONS", "PATH_INFO" => '*')).
+      first.must_equal 200
+
+    lambda do
+      Rack::Lint.new(nil).call(env("PATH_INFO" => "*"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/Only OPTIONS requests may have PATH_INFO set to '\*'/)
+  end
+
+  it "notices request-target authority form errors" do
+    Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "CONNECT", "PATH_INFO" => "example.com:80")).
+      first.must_equal 200
+
+    lambda do
+      Rack::Lint.new(nil).call(env("PATH_INFO" => "example.com:80"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/Only CONNECT requests may have PATH_INFO set to an authority/)
+
+    lambda do
+      Rack::Lint.new(nil).call(env("PATH_INFO" => "/:80")).first.must_equal 200
+    end
+  end
+
+  it "notices request-target absolute-form errors" do
+    Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "GET", "PATH_INFO" => "http://foo/bar")).
+      first.must_equal 200
+
+    lambda do
+      Rack::Lint.new(nil).call(env("REQUEST_METHOD" => "CONNECT", "PATH_INFO" => "http://foo/bar"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/CONNECT and OPTIONS requests must not have PATH_INFO set to a URI/)
+
+    lambda do
+      Rack::Lint.new(nil).call(env("REQUEST_METHOD" => "OPTIONS", "PATH_INFO" => "http://foo/bar"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/CONNECT and OPTIONS requests must not have PATH_INFO set to a URI/)
+  end
+
+  it "notices request-target origin-form errors" do
+    Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "GET", "PATH_INFO" => "/foo/bar")).
+      first.must_equal 200
+
+    lambda do
+      Rack::Lint.new(nil).call(env("REQUEST_METHOD" => "GET", "PATH_INFO" => "../etc/passwd"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/PATH_INFO must start with a '\/'/)
+
+    lambda do
+      Rack::Lint.new(nil).call(env("REQUEST_METHOD" => "GET", "PATH_INFO" => "/foo/bar#qux"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/PATH_INFO.*must not include a fragment/)
+
+    lambda do
+      Rack::Lint.new(nil).call(env("REQUEST_METHOD" => "GET", "PATH_INFO" => "/foo/bar?baz#qux"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/PATH_INFO.*must not include a fragment/)
+  end
+
   it "notice status errors" do
     lambda {
       Rack::Lint.new(lambda { |env|
                        ["cc", {}, ""]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
-      message.must_match(/must be >=100 seen as integer/)
+      message.must_match(/must be an Integer >=100/)
 
     lambda {
       Rack::Lint.new(lambda { |env|
                        [42, {}, ""]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
-      message.must_match(/must be >=100 seen as integer/)
+      message.must_match(/must be an Integer >=100/)
+
+    lambda {
+      Rack::Lint.new(lambda { |env|
+                       ["200", {}, ""]
+                     }).call(env({}))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/must be an Integer >=100/)
   end
 
   it "notice header errors" do
+    obj = Object.new
+    def obj.each; end
     lambda {
       io = StringIO.new('a')
       io.binmode
       Rack::Lint.new(lambda { |env|
                        env['rack.input'].each{ |x| }
-                       [200, Object.new, []]
+                       [200, obj, []]
                      }).call(env({ "rack.input" => io }))
     }.must_raise(Rack::Lint::LintError).
-      message.must_equal "headers object should respond to #each, but doesn't (got Object as headers)"
+      message.must_equal "headers object should be a hash, but isn't (got Object as headers)"
+    lambda {
+      Rack::Lint.new(lambda { |env|
+                       [200, {}.freeze, []]
+                     }).call(env({}))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "headers object should not be frozen, but is"
+
 
     lambda {
       Rack::Lint.new(lambda { |env|
@@ -301,17 +429,18 @@ describe Rack::Lint do
 
     lambda {
       Rack::Lint.new(lambda { |env|
-                       [200, { "Status" => "404" }, []]
+                       [200, { "status" => "404" }, []]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
-      message.must_match(/must not contain Status/)
+      message.must_match(/must not contain status/)
 
     # From RFC 7230:<F24><F25>
     # Most HTTP header field values are defined using common syntax
     # components (token, quoted-string, and comment) separated by
     # whitespace or specific delimiting characters.  Delimiters are chosen
     # from the set of US-ASCII visual characters not allowed in a token
-    # (DQUOTE and "(),/:;<=>?@[\]{}").
+    # (DQUOTE and "(),/:;<=>?@[\]{}"). Rack also doesn't allow uppercase
+    # ASCII (A-Z) in header keys.
     #
     #   token          = 1*tchar
     #
@@ -328,7 +457,15 @@ describe Rack::Lint do
       }.must_raise(Rack::Lint::LintError, "on invalid header: #{invalid_header}").
       message.must_equal("invalid header name: #{invalid_header}")
     end
-    valid_headers = 0.upto(127).map(&:chr) - invalid_headers
+    ('A'..'Z').each do |invalid_header|
+      lambda {
+        Rack::Lint.new(lambda { |env|
+          [200, { invalid_header => "text/plain" }, []]
+        }).call(env({}))
+      }.must_raise(Rack::Lint::LintError, "on invalid header: #{invalid_header}").
+      message.must_equal("uppercase character in header name: #{invalid_header}")
+    end
+    valid_headers = 0.upto(127).map(&:chr) - invalid_headers - ('A'..'Z').to_a
     valid_headers.each do |valid_header|
       Rack::Lint.new(lambda { |env|
                        [200, { valid_header => "text/plain" }, []]
@@ -337,52 +474,94 @@ describe Rack::Lint do
 
     lambda {
       Rack::Lint.new(lambda { |env|
-                       [200, { "Foo" => Object.new }, []]
+                       [200, { "foo" => Object.new }, []]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
-      message.must_equal "a header value must be a String, but the value of 'Foo' is a Object"
+      message.must_equal "a header value must be a String or Array of Strings, but the value of 'foo' is a Object"
 
     lambda {
       Rack::Lint.new(lambda { |env|
-                       [200, { "Foo" => [1, 2, 3] }, []]
-                     }).call(env({}))
-    }.must_raise(Rack::Lint::LintError).
-      message.must_equal "a header value must be a String, but the value of 'Foo' is a Array"
-
-
-    lambda {
-      Rack::Lint.new(lambda { |env|
-                       [200, { "Foo-Bar" => "text\000plain" }, []]
+                       [200, { "foo-bar" => "text\000plain" }, []]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/invalid header/)
 
-    # line ends (010).must_be :allowed in header values.?
-    Rack::Lint.new(lambda { |env|
-                     [200, { "Foo-Bar" => "one\ntwo\nthree", "Content-Length" => "0", "Content-Type" => "text/plain" }, []]
-                   }).call(env({})).first.must_equal 200
+    lambda {
+      Rack::Lint.new(lambda { |env|
+                     [200, [%w(content-type text/plain), %w(content-length 0)], []]
+                     }).call(env({}))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal "headers object should be a hash, but isn't (got Array as headers)"
 
-    # non-Hash header responses.must_be :allowed?
-    Rack::Lint.new(lambda { |env|
-                     [200, [%w(Content-Type text/plain), %w(Content-Length 0)], []]
-                   }).call(env({})).first.must_equal 200
+  end
+
+  it "notice rack.early_hints errors" do
+    def self.env(arg={})
+      super(arg.merge("rack.early_hints" => proc{}))
+    end
+    def self.app(value)
+      app = Rack::Lint.new(lambda { |env|
+                             env['rack.early_hints'].call(value)
+                             [200, {}, []]
+                           })
+      lambda { app.call(env) }
+    end
+
+    app(Object.new).must_raise(Rack::Lint::LintError).
+      message.must_equal "headers object should be a hash, but isn't (got Object as headers)"
+
+    app({}.freeze).must_raise(Rack::Lint::LintError).
+      message.must_equal "headers object should not be frozen, but is"
+
+    app(true => false).must_raise(Rack::Lint::LintError).
+      message.must_equal "header key must be a string, was TrueClass"
+
+    app("status" => "404").must_raise(Rack::Lint::LintError).
+      message.must_match(/must not contain status/)
+
+    invalid_headers = 0.upto(31).map(&:chr) + %W<( ) , / : ; < = > ? @ [ \\ ] { } \x7F>
+    invalid_headers.each do |invalid_header|
+      app(invalid_header => "text/plain").
+        must_raise(Rack::Lint::LintError, "on invalid header: #{invalid_header}").
+        message.must_equal("invalid header name: #{invalid_header}")
+    end
+
+    ('A'..'Z').each do |invalid_header|
+      app(invalid_header => "text/plain").
+        must_raise(Rack::Lint::LintError, "on invalid header: #{invalid_header}").
+        message.must_equal("uppercase character in header name: #{invalid_header}")
+    end
+
+    valid_headers = 0.upto(127).map(&:chr) - invalid_headers - ('A'..'Z').to_a
+    valid_headers.each do |valid_header|
+      app(valid_header => "text/plain").call.first.must_equal 200
+    end
+
+    app("foo" => Object.new).must_raise(Rack::Lint::LintError).
+      message.must_equal "a header value must be a String or Array of Strings, but the value of 'foo' is a Object"
+
+    app("foo-bar" => "text\000plain").must_raise(Rack::Lint::LintError).
+      message.must_match(/invalid header/)
+
+    app([%w(content-type text/plain), %w(content-length 0)]).must_raise(Rack::Lint::LintError).
+      message.must_equal "headers object should be a hash, but isn't (got Array as headers)"
   end
 
   it "notice content-type errors" do
     # lambda {
     #   Rack::Lint.new(lambda { |env|
-    #                    [200, {"Content-length" => "0"}, []]
+    #                    [200, {"content-length" => "0"}, []]
     #                  }).call(env({}))
     # }.must_raise(Rack::Lint::LintError).
-    #   message.must_match(/No Content-Type/)
+    #   message.must_match(/No content-type/)
 
     [100, 101, 204, 304].each do |status|
       lambda {
         Rack::Lint.new(lambda { |env|
-                         [status, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                         [status, { "content-type" => "text/plain", "content-length" => "0" }, []]
                        }).call(env({}))
       }.must_raise(Rack::Lint::LintError).
-        message.must_match(/Content-Type header found/)
+        message.must_match(/content-type header found/)
     end
   end
 
@@ -390,35 +569,186 @@ describe Rack::Lint do
     [100, 101, 204, 304].each do |status|
       lambda {
         Rack::Lint.new(lambda { |env|
-                         [status, { "Content-length" => "0" }, []]
+                         [status, { "content-length" => "0" }, []]
                        }).call(env({}))
       }.must_raise(Rack::Lint::LintError).
-        message.must_match(/Content-Length header found/)
+        message.must_match(/content-length header found/)
     end
 
     lambda {
       Rack::Lint.new(lambda { |env|
-                       [200, { "Content-type" => "text/plain", "Content-Length" => "1" }, []]
+                       [200, { "content-type" => "text/plain", "content-length" => "1" }, []]
                      }).call(env({}))[2].each { }
     }.must_raise(Rack::Lint::LintError).
-      message.must_match(/Content-Length header was 1, but should be 0/)
+      message.must_match(/content-length header was 1, but should be 0/)
+  end
+
+  it "responds to to_path" do
+    body = Object.new
+    def body.each; end
+    def body.to_path; __FILE__ end
+    app = lambda { |env| [200, {}, body] }
+
+    body = Rack::Lint.new(app).call(env({}))[2]
+    body.must_respond_to(:to_path)
+    body.to_path.must_equal __FILE__
   end
 
   it "notice body errors" do
     lambda {
       body = Rack::Lint.new(lambda { |env|
-                               [200, { "Content-type" => "text/plain", "Content-length" => "3" }, [1, 2, 3]]
+                               [200, { "content-type" => "text/plain", "content-length" => "3" }, [1, 2, 3]]
                              }).call(env({}))[2]
       body.each { |part| }
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/yielded non-string/)
+
+    lambda {
+      body = Rack::Lint.new(lambda { |env|
+                               [200, { "content-type" => "text/plain", "content-length" => "3" }, Object.new]
+                             }).call(env({}))[2]
+      body.respond_to?(:to_ary).must_equal false
+      body.each { |part| }
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'Enumerable Body must respond to each'
+
+    lambda {
+      body = Rack::Lint.new(lambda { |env|
+                               [200, { "content-type" => "text/plain", "content-length" => "0" }, []]
+                             }).call(env({}))[2]
+      body.each { |part| }
+      body.each { |part| }
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'Response body must only be invoked once (each)'
+
+    # Lint before and after the Rack middleware being tested.
+    def stacked_lint(app)
+      Rack::Lint.new(lambda do |env|
+        Rack::Lint.new(app).call(env).tap {|response| response[2] = yield response[2]}
+      end)
+    end
+
+    yielder_app = lambda do |_|
+      input = Object.new
+      def input.each; 10.times {yield 'foo'}; end
+      [200, {"content-type" => "text/plain", "content-length" => "30"}, input]
+    end
+
+    lambda {
+      body = stacked_lint(yielder_app) {|body|
+        new_body = Struct.new(:body) do
+          def each(&block)
+            body.each { |part| yield part.upcase }
+            body.close
+          end
+        end
+        new_body.new(body)
+      }.call(env({}))[2]
+      body.each {|part| part.must_equal 'FOO'}
+      body.close
+    }.call
+
+    lambda {
+      body = stacked_lint(yielder_app) { |body|
+        body.enum_for.to_a
+      }.call(env({}))[2]
+      body.each {}
+      body.close
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/Middleware must not call #each directly/)
+
+    lambda {
+      body = stacked_lint(yielder_app) { |body|
+        new_body = Struct.new(:body) do
+          def each(&block)
+            body.enum_for.each_slice(2) { |parts| yield parts.join }
+          end
+        end
+        new_body.new(body)
+      }.call(env({}))[2]
+      body.each {}
+      body.close
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/New body must yield at least once per iteration of old body/)
+
+    lambda {
+      body = stacked_lint(yielder_app) { |body|
+        Struct.new(:body) do
+          def each; body.each {|part| yield part} end
+        end.new(body)
+      }.call(env({}))[2]
+      body.each {}
+      body.close
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/Body has not been closed/)
+
+    static_app = lambda do |_|
+      input = ['foo'] * 10
+      [200, {"content-type" => "text/plain", "content-length" => "30"}, input]
+    end
+
+    lambda {
+      body = stacked_lint(static_app) { |body| body.to_ary}.call(env({}))[2]
+      body.each {}
+      body.close
+    }.call
+
+    array_mismatch = lambda do |_|
+      input = Object.new
+      def input.to_ary; ['bar'] * 10; end
+      def input.each; 10.times {yield 'foo'}; end
+      [200, {"content-type" => "text/plain", "content-length" => "30"}, input]
+    end
+
+    lambda {
+      body = stacked_lint(array_mismatch) { |body| body.to_ary}.call(env({}))[2]
+      body.each {}
+      body.close
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/#to_ary not identical to contents produced by calling #each/)
+
+    lambda {
+      body = Rack::Lint.new(lambda { |env|
+                               to_path = Object.new
+                               def to_path.each; end
+                               def to_path.to_path; 'non-existent' end
+                               [200, { "content-type" => "text/plain", "content-length" => "0" }, to_path]
+                             }).call(env({}))[2]
+      body.each { |part| }
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'The file identified by body.to_path does not exist'
+
+    lambda {
+      body = Rack::Lint.new(lambda { |env|
+                               [200, { "content-type" => "text/plain", "content-length" => "0" }, Object.new]
+                             }).call(env({}))[2]
+      body.call(nil)
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'Streaming Body must respond to call'
+
+    lambda {
+      body = Rack::Lint.new(lambda { |env|
+                               [200, { "content-type" => "text/plain", "content-length" => "0" }, proc{}]
+                             }).call(env({}))[2]
+      body.call(StringIO.new)
+      body.call(nil)
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'Response body must only be invoked once (call)'
+
+    lambda {
+      body = Rack::Lint.new(lambda { |env|
+                               [200, { "content-type" => "text/plain", "content-length" => "0" }, proc{}]
+                             }).call(env({}))[2]
+      body.call(nil)
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'Stream must respond to read'
   end
 
   it "notice input handling errors" do
     lambda {
       Rack::Lint.new(lambda { |env|
                        env["rack.input"].gets("\r\n")
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/gets called with arguments/)
@@ -427,7 +757,7 @@ describe Rack::Lint do
       Rack::Lint.new(lambda { |env|
                        env["rack.input"].gets
                        env["rack.input"].read(1, 2, 3)
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/read called with too many arguments/)
@@ -435,7 +765,7 @@ describe Rack::Lint do
     lambda {
       Rack::Lint.new(lambda { |env|
                        env["rack.input"].read("foo")
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/read called with non-integer and non-nil length/)
@@ -443,7 +773,7 @@ describe Rack::Lint do
     lambda {
       Rack::Lint.new(lambda { |env|
                        env["rack.input"].read(-1)
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/read called with a negative length/)
@@ -451,7 +781,7 @@ describe Rack::Lint do
     lambda {
       Rack::Lint.new(lambda { |env|
                        env["rack.input"].read(nil, nil)
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/read called with non-String buffer/)
@@ -459,19 +789,11 @@ describe Rack::Lint do
     lambda {
       Rack::Lint.new(lambda { |env|
                        env["rack.input"].read(nil, 1)
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/read called with non-String buffer/)
 
-    lambda {
-      Rack::Lint.new(lambda { |env|
-                       env["rack.input"].rewind(0)
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
-                     }).call(env({}))
-    }.must_raise(Rack::Lint::LintError).
-      message.must_match(/rewind called with arguments/)
-
     weirdio = Object.new
     class << weirdio
       def gets
@@ -486,10 +808,6 @@ describe Rack::Lint do
         yield 23
         yield 42
       end
-
-      def rewind
-        raise Errno::ESPIPE, "Errno::ESPIPE"
-      end
     end
 
     eof_weirdio = Object.new
@@ -504,23 +822,28 @@ describe Rack::Lint do
 
       def each
       end
-
-      def rewind
-      end
     end
 
     lambda {
       Rack::Lint.new(lambda { |env|
                        env["rack.input"].gets
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                      }).call(env("rack.input" => weirdio))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/gets didn't return a String/)
 
+    lambda {
+      Rack::Lint.new(lambda { |env|
+                       env["rack.input"].each(1) { |x| }
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
+                     }).call(env("rack.input" => weirdio))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/rack.input#each called with arguments/)
+
     lambda {
       Rack::Lint.new(lambda { |env|
                        env["rack.input"].each { |x| }
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                      }).call(env("rack.input" => weirdio))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/each didn't yield a String/)
@@ -528,7 +851,7 @@ describe Rack::Lint do
     lambda {
       Rack::Lint.new(lambda { |env|
                        env["rack.input"].read
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                      }).call(env("rack.input" => weirdio))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/read didn't return nil or a String/)
@@ -536,34 +859,28 @@ describe Rack::Lint do
     lambda {
       Rack::Lint.new(lambda { |env|
                        env["rack.input"].read
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                      }).call(env("rack.input" => eof_weirdio))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/read\(nil\) returned nil on EOF/)
+  end
 
-    lambda {
-      Rack::Lint.new(lambda { |env|
-                       env["rack.input"].rewind
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
-                     }).call(env("rack.input" => weirdio))
-    }.must_raise(Rack::Lint::LintError).
-      message.must_match(/rewind raised Errno::ESPIPE/)
+  it "can call close" do
+    app = lambda do |env|
+      env["rack.input"].close
+      [201, {"content-type" => "text/plain", "content-length" => "0"}, []]
+    end
 
+    response = Rack::Lint.new(app).call(env({}))
 
-    lambda {
-      Rack::Lint.new(lambda { |env|
-                       env["rack.input"].close
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
-                     }).call(env({}))
-    }.must_raise(Rack::Lint::LintError).
-      message.must_match(/close must not be called/)
+    response.first.must_equal 201
   end
 
   it "notice error handling errors" do
     lambda {
       Rack::Lint.new(lambda { |env|
                        env["rack.errors"].write(42)
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/write not called with a String/)
@@ -571,7 +888,7 @@ describe Rack::Lint do
     lambda {
       Rack::Lint.new(lambda { |env|
                        env["rack.errors"].close
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                      }).call(env({}))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/close must not be called/)
@@ -579,12 +896,12 @@ describe Rack::Lint do
 
   it "notice HEAD errors" do
     Rack::Lint.new(lambda { |env|
-                     [200, { "Content-type" => "test/plain", "Content-length" => "3" }, []]
+                     [200, { "content-type" => "test/plain", "content-length" => "3" }, []]
                    }).call(env({ "REQUEST_METHOD" => "HEAD" })).first.must_equal 200
 
     lambda {
       Rack::Lint.new(lambda { |env|
-                       [200, { "Content-type" => "test/plain", "Content-length" => "3" }, ["foo"]]
+                       [200, { "content-type" => "test/plain", "content-length" => "3" }, ["foo"]]
                      }).call(env({ "REQUEST_METHOD" => "HEAD" }))[2].each { }
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/body was given for HEAD/)
@@ -596,7 +913,7 @@ describe Rack::Lint do
 
     Rack::Lint.new(lambda { |env|
                      env["rack.input"].send(:read, *args)
-                     [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
+                     [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
                    }).call(env({ "rack.input" => StringIO.new(hello_str) })).
       first.must_equal 201
   end
@@ -610,36 +927,67 @@ describe Rack::Lint do
     assert_lint 1, ''.dup
   end
 
-  it "notice hijack errors" do
+  it "notices when request env doesn't have a valid rack.hijack callback" do
     lambda {
       Rack::Lint.new(lambda { |env|
                        env['rack.hijack'].call
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
-                     }).call(env({ 'rack.hijack?' => true, 'rack.hijack' => lambda { Object.new } }))
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
+                     }).call(env({ 'rack.hijack' => Object.new }))
     }.must_raise(Rack::Lint::LintError).
-      message.must_match(/rack.hijack_io must respond to read/)
-
-      Rack::Lint.new(lambda { |env|
-                       env['rack.hijack'].call
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0" }, []]
-                     }).call(env({ 'rack.hijack?' => true, 'rack.hijack' => lambda { StringIO.new }, 'rack.hijack_io' => StringIO.new })).
-        first.must_equal 201
-
-      Rack::Lint.new(lambda { |env|
-                       env['rack.hijack?'] = true
-                       [201, { "Content-type" => "text/plain", "Content-length" => "0", 'rack.hijack' => lambda {|io| io }, 'rack.hijack_io' => StringIO.new }, []]
-                     }).call(env({}))[1]['rack.hijack'].call(StringIO.new).read.must_equal ''
+      message.must_match(/rack.hijack must respond to call/)
   end
 
-end
+  it "notices when the response headers don't have a valid rack.hijack callback" do
+    lambda {
+      Rack::Lint.new(lambda { |env|
+                       [201, { "content-type" => "text/plain", "content-length" => "0", 'rack.hijack' =>  Object.new }, []]
+                     }).call(env({ 'rack.hijack?' => true }))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'rack.hijack header must respond to #call'
+  end
 
-describe "Rack::Lint::InputWrapper" do
-  it "delegate :rewind to underlying IO object" do
-    io = StringIO.new("123")
-    wrapper = Rack::Lint::InputWrapper.new(io)
-    wrapper.read.must_equal "123"
-    wrapper.read.must_equal ""
-    wrapper.rewind
-    wrapper.read.must_equal "123"
+  it "pass valid rack.response_finished" do
+    callable_object = Class.new do
+      def call(env, status, headers, error)
+      end
+    end.new
+
+    Rack::Lint.new(lambda { |env|
+                     [200, {}, ["foo"]]
+                   }).call(env({ "rack.response_finished" => [-> (env) {}, lambda { |env| }, callable_object], "content-length" => "3" })).first.must_equal 200
+  end
+
+  it "notices when the response protocol is specified in the response but not in the request" do
+    app = Rack::Lint.new(lambda{|env|
+      [101, {'rack.protocol' => 'websocket'}, ["foo"]]
+    })
+
+    lambda do
+      app.call(env())
+    end
+      .must_raise(Rack::Lint::LintError)
+      .message.must_match(/rack.protocol header is "websocket", but rack.protocol was not set in request/)
+  end
+
+  it "notices when the response protocol is specified in the response but not in the request" do
+    app = Rack::Lint.new(lambda{|env|
+      [101, {'rack.protocol' => 'websocket'}, ["foo"]]
+    })
+
+    lambda do
+      app.call(env('rack.protocol' => ['smtp']))
+    end
+      .must_raise(Rack::Lint::LintError)
+      .message.must_match(/rack.protocol header is "websocket", but should be one of \["smtp"\] from the request!/)
+  end
+
+  it "pass valid rack.protocol" do
+    app = Rack::Lint.new(lambda{|env|
+      [101, {'rack.protocol' => 'websocket'}, ["foo"]]
+    })
+
+    response = app.call(env({'rack.protocol' => ['websocket']}))
+
+    response.first.must_equal 101
   end
 end
